@@ -16,7 +16,9 @@ use Class::MethodMaker
 	     'auth_user',
 	     'auth_passwd',
 	     'override_locks',
+	     'release_edit_lock',
 	     'verbose',
+             'skin_hints',
             ],
  new_hash_init => 'hash_init'
  ;
@@ -33,7 +35,9 @@ sub new {
 sub pre_init {
   my $self   = shift;
   $self->override_locks (0);
+  $self->release_edit_lock (1);
   $self->verbose (0);
+  $self->skin_hints ({});
 }
 
 sub post_init {
@@ -49,7 +53,7 @@ sub get_basic_credentials {
 
 # constructs URL
 # if topic doesn't contain a Web prefix, "current_default_web" is prepended
-sub make_url {
+sub _make_url {
   my $self  = shift;
   my $cmd   = shift;
   my $topic = shift;
@@ -68,23 +72,30 @@ sub make_url {
   return $url;
 }
 
-sub skin_regex_topic_locked {
+sub _skin_regex_topic_locked {
   my $self = shift;
-  return qr/name="Topic_is_locked_by_another_user"/;
+  return qr/\(oops\).*name="Topic_is_locked_by_another_user"/s;
 }
 
-sub skin_regex_topic_locked_edit_anyway {
+sub _skin_regex_topic_locked_edit_anyway {
   my $self = shift;
   return qr/Edit anyway/;
 }
 
-sub skin_regex_authentication_failed {
+sub _skin_regex_authentication_failed {
   my $self = shift;
-  return qr/name="Either_you_need_to_register_or_t"/;
+  return qr/TWikiRegistration.*\(oops\).*name="Either_you_need_to_register_or_t"/s;
+}
+
+sub _skin_regex_save_or_preview_page {
+  my $self = shift;
+  my $topic = shift || ''; # needed for "where I am"-heuristic
+
+  return qr/form name=".*".*action=".*\/save\/.*$topic">/s;
 }
 
 # a little helper function
-sub htmplparse_get_text {
+sub _htmplparse_get_text {
   my $self = shift;
 
   my($p, $stop) = @_;
@@ -121,7 +132,7 @@ sub htmlparse_extract_single_textarea {
 	next unless ref $t;  # skip text
 	last if $t->[0] eq "/form";
 	if ($t->[0] eq "textarea") {
-	  return $self->htmplparse_get_text ($p, "/textarea");
+	  return $self->_htmplparse_get_text ($p, "/textarea");
 	}
       }
     }
@@ -132,54 +143,94 @@ sub htmlparse_extract_single_textarea {
 sub edit_press_cancel {
   my $self = shift;
 
-  my $url = $self->make_url ('view', $self->current_topic, '?unlock=on');
+  my $url = $self->_make_url ('view', $self->current_topic, '?unlock=on');
   #print STDERR "edit_press_cancel: $url\n" if $self->verbose;
   $self->follow_link (url => $url);
 }
 
-
 sub read_topic {
   my $self = shift;
   my $topic = shift || $self->current_topic;
-  my $url = $self->make_url ('view', $topic, '?raw=on');
+  my $url = $self->_make_url ('view', $topic, '?raw=on');
   #print STDERR "read_topic: $url\n" if $self->verbose;
   $self->get ($url);
   return $self->htmlparse_extract_single_textarea ($self->content);
 }
 
+sub _handle_release_edit_lock {
+  my $self = shift;
+
+  my $unlock_checkbox = $self->current_form->find_input ('unlock', 'checkbox');
+  # "release edit lock"
+  if ($unlock_checkbox) {
+    if ($self->release_edit_lock) {
+      $self->tick ('unlock', 'on');
+    } else {
+      $self->untick ('unlock', 'on');
+    }
+  }
+}
+
 sub save_topic {
   my $self = shift;
   my $content = shift;
   my $topic = shift || $self->current_topic;
 
-  my $url = $self->make_url ('edit', $topic);
+  my $url = $self->_make_url ('edit', $topic);
   #print STDERR "save_topic: $url\n" if $self->verbose;
 
   # get page
   $self->get ($url);
 
   # locked?
-  my $html_content = $self->content;
-  if ($html_content =~ $self->skin_regex_topic_locked) {
-    if ($self->override_locks) {
-      # edit anyway
-      $self->follow_link (text_regex => $self->skin_regex_topic_locked_edit_anyway);
-      $self->get ($url);
-    } else {
-      print STDERR "Topic is locked.\n" if $self->verbose;
-      return undef;
-    }
-  } elsif ($html_content =~ $self->skin_regex_authentication_failed) {
-    print STDERR "Access denied. Authentication failed.\n" if $self->verbose;
-    return undef;
-  }
+  $self->_save_topic_handle_locks ($url) or return undef;
 
   # fill form
   $self->form_number (1);
   $self->current_form;
   $self->set_fields ( text => $content );
-  $self->click_button ( value => "Save" );
+  $self->_save_topic_Save ($topic);
   return 1;
 }
 
+sub _save_topic_handle_locks {
+  my $self  = shift;
+  my $url   = shift;
+
+  my $html_content = $self->content;
+  if ($html_content =~ $self->_skin_regex_topic_locked) {
+    if ($self->override_locks) {
+      # edit anyway
+      print STDERR "Override topic lock.\n" if $self->verbose;
+      $self->follow_link (text_regex => $self->_skin_regex_topic_locked_edit_anyway);
+      $self->get ($url);
+    } else {
+      print STDERR "Topic is locked.\n" if $self->verbose;
+      return undef;
+    }
+  } elsif ($html_content =~ $self->_skin_regex_authentication_failed) {
+    print STDERR "Access denied. Authentication failed.\n" if $self->verbose;
+    return undef;
+  }
+  return 1;
+}
+
+sub _save_topic_Save {
+  my $self  = shift;
+  my $topic = shift || ''; # needed for "where I am"-heuristic
+
+  $self->_handle_release_edit_lock;
+  # simply submit (== either "Preview Changes" or "Save Changes")
+  $self->submit();
+  # did we arrive at a preview page?
+  my $content = $self->content;
+  if ($content =~ _skin_regex_save_or_preview_page ($topic)) {
+    # simply submit again (== "Save Changes")
+    $self->_handle_release_edit_lock;
+    $self->submit();
+  }
+}
+
 1;
+
+
